from typing import Optional, Generic, TypeVar, List, Union, Type, Dict
import os

T = TypeVar('T')


class ConfigDict(Dict[str, Union[str, int, float, bool]]):

  @classmethod
  def from_string(cls, value: str) -> 'ConfigDict':
    result = cls()
    if not value:
      return result
    pairs = value.split(',')
    for pair in pairs:
      key, val = pair.split('=')
      key = key.strip()
      val = val.strip()
      # Try to parse the value as int, float, or bool
      if val.lower() == 'true':
        result[key] = True
      elif val.lower() == 'false':
        result[key] = False
      else:
        try:
          result[key] = int(val)
        except ValueError:
          try:
            result[key] = float(val)
          except ValueError:
            result[key] = val
    return result


class StrList(List[str]):

  @classmethod
  def from_string(cls, value: str) -> 'StrList':
    return cls(item.strip() for item in value.split(';') if item.strip()
              ) if value.strip() else cls()


class IntList(List[int]):

  @classmethod
  def from_string(cls, value: str) -> 'IntList':
    return cls(int(item.strip()) for item in value.split(';') if item.strip()
              ) if value.strip() else cls()


class FloatList(List[float]):

  @classmethod
  def from_string(cls, value: str) -> 'FloatList':
    return cls(
      float(item.strip()) for item in value.split(';') if item.strip()
    ) if value.strip() else cls()


class BoolList(List[bool]):

  @classmethod
  def from_string(cls, value: str) -> 'BoolList':
    return cls(
      item.strip().lower() == 'true'
      for item in value.split(';')
      if item.strip()
    ) if value.strip() else cls()


class Config(Generic[T]):
  name: str
  type: Type[T]
  default: str
  description: Optional[str]
  __value__: T

  def __init__(
    self,
    name: str,
    type: Type[T],
    default: str,
    description: Optional[str] = None
  ):
    self.name = name
    self.type = type
    self.default = default
    self.description = description
    self.__value__ = self.resolve_value()

  @property
  def value(self) -> T:
    return self.__value__

  def resolve_value(self) -> Union[T, List[T]]:
    if '*' in self.name:
      return self._resolve_wildcard()
    else:
      return self._resolve_single()

  def _resolve_single(self) -> T:
    raw_value = os.getenv(self.name, self.default)
    if isinstance(raw_value, list):
      raw_value = raw_value[0] if raw_value else ''
    return self._convert_value(raw_value)

  def _resolve_wildcard(self) -> List[T]:
    prefix = self.name.replace('*', '')
    matching_vars = [
      (key, value)
      for key, value in os.environ.items()
      if key.startswith(prefix)
    ]

    if not matching_vars:
      if isinstance(self.default, str):
        return [self._convert_value(self.default)] if self.default else []
      return self.default

    return [self._convert_value(value) for _, value in sorted(matching_vars)]

  def _convert_value(self, value: str) -> T:
    if issubclass(
      self.type, (StrList, IntList, FloatList, BoolList, ConfigDict)
    ):
      return self.type.from_string(value)
    elif self.type == str:
      return value
    elif self.type == int:
      return int(value)
    elif self.type == float:
      return float(value)
    elif self.type == bool:
      return value.lower() in ('true', '1', 'yes', 'on')
    else:
      return self.type(value)


# ----------------- APIs -----------------

HARBOR_OPENAI_URLS = Config[StrList](
  name='HARBOR_OPENAI_URLS',
  type=StrList,
  default='',
  description='A list of URLs to the OpenAI APIs'
)

HARBOR_OPENAI_KEYS = Config[StrList](
  name='HARBOR_OPENAI_KEYS',
  type=StrList,
  default='',
  description='A list of API keys to use for the OpenAI APIs'
)

HARBOR_BOOST_OPENAI_URLS = Config[StrList](
  name='HARBOR_BOOST_OPENAI_URLS',
  type=StrList,
  default='',
  description='A list of URLs to the OpenAI APIs to boost'
)

HARBOR_BOOST_OPENAI_KEYS = Config[StrList](
  name='HARBOR_BOOST_OPENAI_KEYS',
  type=StrList,
  default='',
  description='A list of API keys to use for the OpenAI APIs to boost'
)

HARBOR_BOOST_EXTRA_OPENAI_URLS = Config[str](
  name='HARBOR_BOOST_OPENAI_URL_*',
  type=str,
  default='',
  description='Named OpenAI-compatible API URLs to boost'
)

HARBOR_BOOST_EXTRA_OPENAI_KEYS = Config[str](
  name='HARBOR_BOOST_OPENAI_KEY_*',
  type=str,
  default='',
  description=
  'Named OpenAI-compatible API keys to use for the OpenAI APIs to boost'
)

# Combining all the sources from
# above into a single list
BOOST_APIS = [
  *HARBOR_OPENAI_URLS.value, *HARBOR_BOOST_OPENAI_URLS.value,
  *HARBOR_BOOST_EXTRA_OPENAI_URLS.value
]

BOOST_KEYS = [
  *HARBOR_OPENAI_KEYS.value, *HARBOR_BOOST_OPENAI_KEYS.value,
  *HARBOR_BOOST_EXTRA_OPENAI_KEYS.value
]

# ----------------- MODULES -----------------

BOOST_MODS = Config[StrList](
  name='HARBOR_BOOST_MODULES',
  type=StrList,
  default='klmbr;rcn;g1',
  description='A list of boost modules to load'
)

BOOST_FOLDERS = Config[StrList](
  name='HARBOR_BOOST_MODULE_FOLDERS',
  type=StrList,
  default='modules;custom_modules',
  description='A list of folders to load boost modules from'
)

# ---------------- COMPLETION ---------------

INTERMEDIATE_OUTPUT = Config[bool](
  name='HARBOR_BOOST_INTERMEDIATE_OUTPUT',
  type=bool,
  default='true',
  description='Whether to output intermediate completions'
)

STATUS_STYLE = Config[str](
  name='HARBOR_BOOST_STATUS_STYLE',
  type=str,
  default='md:codeblock',
  description='The style of status messages'
)

# ---------------- BEHAVIOR -----------------

SERVE_BASE_MODELS = Config[bool](
  name='HARBOR_BOOST_BASE_MODELS',
  type=bool,
  default='false',
  description=
  'When enabled, boost will also serve original models from downstream APIs'
)

MODEL_FILTER = Config[ConfigDict](
  name='HARBOR_BOOST_MODEL_FILTER',
  type=ConfigDict,
  default='',
  description=
  'When specified, boost will only serve models which IDs match the filter'
)

API_KEY = Config[str](
  name='HARBOR_BOOST_API_KEY',
  type=str,
  default='',
  description='The API key to use for the boost API'
)

API_KEYS = Config[StrList](
  name='HARBOR_BOOST_API_KEYS',
  type=StrList,
  default='',
  description='A colon-separated list of API keys to use for the boost API'
)

EXTRA_KEYS = Config[str](
  name='HARBOR_BOOST_API_KEY_*',
  type=str,
  default='',
  description='Named API keys to use for the boost API'
)

BOOST_AUTH = [
  key for key in [API_KEY.value, *API_KEYS.value, *EXTRA_KEYS.value] if key
]

# ------------------ KLMBR ------------------

KLMBR_PERCENTAGE = Config[int](
  name='HARBOR_BOOST_KLMBR_PERCENTAGE',
  type=int,
  default='15',
  description='The percentage of text to modify with the klmbr module'
)

KLMBR_MODS = Config[StrList](
  name='HARBOR_BOOST_KLMBR_MODS',
  type=StrList,
  default='',
  description=f'The list of modifications klmbr will apply'
)

KLMBR_STRAT = Config[str](
  name='HARBOR_BOOST_KLMBR_STRAT',
  type=str,
  default='all',
  description='The strategy that selects messages to modify for the klmbr module'
)

KLMBR_STRAT_PARAMS = Config[ConfigDict](
  name='HARBOR_BOOST_KLMBR_STRAT_PARAMS',
  type=ConfigDict,
  default='',
  description=
  'The parameters for the strategy that selects messages to modify for the klmbr module'
)

# ----------------- RCN -----------------

RCN_STRAT = Config[str](
  name='HARBOR_RCN_STRAT',
  type=str,
  default='match',
  description='The strategy that selects messages to modify for the rcn module'
)

RCN_STRAT_PARAMS = Config[ConfigDict](
  name='HARBOR_RCN_STRAT',
  type=ConfigDict,
    # Default - last user message
  default='role=user,index=-1',
  description='Parameters for rcn message selection'
)

# ----------------- G1 -----------------

G1_STRAT = Config[str](
  name='HARBOR_G1_STRAT',
  type=str,
  default='match',
  description='The strategy that selects messages to modify for the g1 module'
)

G1_STRAT_PARAMS = Config[ConfigDict](
  name='HARBOR_G1_STRAT_PARAMS',
  type=ConfigDict,
    # Default - last user message
  default='role=user,index=-1',
  description='Parameters for g1 message selection'
)

G1_MAX_STEPS = Config[int](
  name='HARBOR_G1_MAX_STEPS',
  type=int,
  default='15',
  description='The maximum number of reasoning steps to generate'
)

# ----------------- MCTS -----------------

MCTS_STRAT = Config[str](
  name='HARBOR_MCTS_STRAT',
  type=str,
  default='match',
  description='The strategy that selects messages to target for the mcts module'
)

MCTS_STRAT_PARAMS = Config[ConfigDict](
  name='HARBOR_MCTS_STRAT_PARAMS',
  type=ConfigDict,
    # Default - last user message
  default='role=user,index=-1',
  description='Parameters for mcts message selection'
)

MCTS_MAX_SIMULATIONS = Config[int](
  name='HARBOR_MCTS_MAX_SIMULATIONS',
  type=int,
  default='2',
  description='The maximum number of simulations to run (per iteration)'
)

MCTS_MAX_ITERATIONS = Config[int](
  name='HARBOR_MCTS_MAX_ITERATIONS',
  type=int,
  default='2',
  description='The maximum number of iterations to run'
)

MCTS_THOUGHTS = Config[int](
  name='HARBOR_MCTS_THOUGHTS',
  type=int,
  default='2',
  description=
  'The amount of thoughts (node expansions) to generate per simulation'
)

MCTS_EXPLORATION_CONSTANT = Config[float](
  name='HARBOR_MCTS_EXPLORATION_CONSTANT',
  type=float,
  default='1.414',
  description='The exploration constant for the MCTS algorithm'
)

# ----------------- ELI5 -----------------

ELI5_STRAT = Config[str](
  name='HARBOR_ELI5_STRAT',
  type=str,
  default='match',
  description='The strategy that selects messages to target for the eli5 module'
)

ELI5_STRAT_PARAMS = Config[ConfigDict](
  name='HARBOR_ELI5_STRAT_PARAMS',
  type=ConfigDict,
    # Default - last user message
  default='role=user,index=-1',
  description='Parameters for eli5 message selection'
)

# ----------- SUPERSUMMER ----------------

SUPERSUMMER_STRAT = Config[str](
  name='HARBOR_SUPERSUMMER_STRAT',
  type=str,
  default='match',
  description=
  'The strategy that selects messages to target for the supersummer module'
)

SUPERSUMMER_STRAT_PARAMS = Config[ConfigDict](
  name='HARBOR_SUPERSUMMER_STRAT_PARAMS',
  type=ConfigDict,
    # Default - last user message
  default='role=user,index=-1',
  description='Parameters for supersummer message selection'
)

SUPERSUMMER_NUM_QUESTIONS = Config[int](
  name='HARBOR_SUPERSUMMER_NUM_QUESTIONS',
  type=int,
  default='5',
  description='The number of questions to generate for the summarisation'
)

SUPERSUMMER_LENGTH = Config[str](
  name='HARBOR_BOOST_SUPERSUMMER_LENGTH',
  type=str,
  default='few paragraphs',
  description='Desired length of the summary'
)



import httpx
import json

from fastapi import FastAPI, Request, HTTPException, Depends, Security
from fastapi.security.api_key import APIKeyHeader
from fastapi.responses import JSONResponse, StreamingResponse

from config import MODEL_FILTER, SERVE_BASE_MODELS, BOOST_AUTH
from log import setup_logger

import selection
import mapper
import config
import mods
import llm

logger = setup_logger(__name__)
app = FastAPI()
auth_header = APIKeyHeader(name="Authorization", auto_error=False)

# ------------------------------
async def get_api_key(api_key_header: str = Security(auth_header)):
  if len(BOOST_AUTH) == 0:
    return

  if api_key_header is not None:
    # Bearer/plain versions
    value = api_key_header.replace("Bearer ", "").replace("bearer ", "")
    if value in BOOST_AUTH:
      return value

  raise HTTPException(status_code=403, detail="Unauthorized")


@app.get("/")
async def root():
  return JSONResponse(
    content={
      "status": "ok",
      "message": "Harbor Boost is running"
    },
    status_code=200
  )


@app.get("/health")
async def health():
  return JSONResponse(content={"status": "ok"}, status_code=200)


@app.get("/v1/models")
async def get_boost_models(api_key: str = Depends(get_api_key)):
  downstream = await mapper.list_downstream()
  enabled_modules = config.BOOST_MODS.value
  should_filter = len(MODEL_FILTER.value) > 0
  serve_base_models = SERVE_BASE_MODELS.value
  candidates = []
  final = []

  for model in downstream:
    if serve_base_models:
      candidates.append(model)

    for module in enabled_modules:
      mod = mods.registry.get(module)
      candidates.append(mapper.get_proxy_model(mod, model))

  for model in candidates:
    should_serve = True

    if should_filter:
      should_serve = selection.matches_filter(model, MODEL_FILTER.value)
      print(model['id'], MODEL_FILTER.value['id.regex'], should_serve)

    if should_serve:
      final.append(model)

  logger.debug(f"Serving {len(final)} models in the API")

  return JSONResponse(content=final, status_code=200)


@app.post("/v1/chat/completions")
async def post_boost_chat_completion(request: Request, api_key: str = Depends(get_api_key)):
  body = await request.body()

  logger.debug(f"Request body: {body[:256]}...")

  try:
    decoded = body.decode("utf-8")
    json_body = json.loads(decoded)
    stream = json_body.get("stream", False)
  except json.JSONDecodeError:
    logger.debug(f"Invalid JSON in request body: {body[:100]}")
    raise HTTPException(status_code=400, detail="Invalid JSON in request body")

  # Refresh downstream models to ensure
  # that we know where to route the requests
  await mapper.list_downstream()

  # Get our proxy model configuration
  proxy_config = mapper.resolve_request_config(json_body)
  proxy = llm.LLM(**proxy_config)

  # We don't want to trigger potentially
  # expensive workflows for title generation
  if mapper.is_title_generation_task(proxy):
    logger.debug("Detected title generation task, skipping boost")
    return JSONResponse(content=await proxy.chat_completion(), status_code=200)

  # This is where the "boost" happens
  completion = await proxy.serve()

  if completion is None:
    return JSONResponse(
      content={"error": "No completion returned"}, status_code=500
    )

  if stream:
    return StreamingResponse(completion, media_type="text/event-stream")
  else:
    content = await proxy.consume_stream(completion)
    return JSONResponse(content=content, status_code=200)

# ------------ Startup ----------------

logger.info(f"Boosting: {config.BOOST_APIS}")
if len(BOOST_AUTH) == 0:
  logger.warn("No API keys specified - boost will accept all requests")

if __name__ == "__main__":
  import uvicorn
  uvicorn.run(app, host="0.0.0.0", port=8000)



# Recursive Certainty Validation - RCN
# aka "Are you sure?

import re
from typing import Optional, List
from config import MCTS_STRAT, MCTS_STRAT_PARAMS, MCTS_EXPLORATION_CONSTANT, MCTS_MAX_SIMULATIONS, MCTS_MAX_ITERATIONS, MCTS_THOUGHTS

import llm
import log
import random
import math
import chat as ch
import selection

# ==============================================================================

logger = log.setup_logger(__name__)
ID_PREFIX = "mcts"

# ==============================================================================

thoughts_prompt = """
<instruction>
Give a suggestion on how this answer can be improved.
WRITE ONLY AN IMPROVEMENT SUGGESTION AND NOTHING ELSE.
YOUR REPLY SHOULD BE A SINGLE SENTENCE.
</instruction>

<question>
{question}
</question>

<draft>
{answer}
</draft>
""".strip()

eval_answer_prompt = """
Given the following text:
"{answer}"

How well does it answers this question:
"{question}"

Rate the answer from 1 to 10, where 1 is completely wrong or irrelevant and 10 is a perfect answer.
Reply with a single number between 1 and 10 only. Do not write anything else, it will be discarded.
THINK CAREFULLY AND USE BEST PRACTICES.
""".strip()

analyze_prompt = """
Iteration Analysis:

Original question: {question}
Best answer found: {best_answer}
Best score achieved: {best_score}

Analyze this iteration of the thought process. Consider the following:
1. What aspects of the best answer made it successful?
2. What patterns or approaches led to higher-scoring thoughts?
3. Were there any common pitfalls or irrelevant tangents in lower-scoring thoughts?
4. How can the thought generation process be improved for the next iteration?

Provide a concise analysis and suggest one specific improvement strategy for the next iteration.
""".strip()

update_prompt = """
<instruction>
Your task is to read the question and the answer below, then analyse the given critique.
When you are done - think about how the answer can be improved based on the critique.
WRITE A REVISED ANSWER THAT ADDRESSES THE CRITIQUE. DO NOT WRITE ANYTHING ELSE.
</instruction>
<question>
{question}
</question>
<draft>
{answer}
</draft>
<critique>
{improvements}
</critique>
""".strip()

initial_prompt = """
<instruction>
Answer the question below. Do not pay attention to, unexpected casing, punctuation or accent marks.
</instruction>

<question>
{question}
</question>
"""


class MCTSNode(ch.ChatNode):
  children: List['MCTSNode']
  exploration_weight: float
  max_children = 2

  def fully_expanded(self):
    return len(self.children) >= self.max_children

  def uct_value(self):
    epsilon = 1e-6

    return self.value / (self.visits +
                         epsilon) + self.exploration_weight * math.sqrt(
                           math.log(self.parent.visits) /
                           (self.visits + epsilon)
                         )

  def mermaid(self, offset=0, selected=None):
    padding = " " * offset
    msg = f"{padding}{self.id}({self.id}:{self.visits} - {escape_mermaid(self.content[:25])})\n"

    if selected == self.id:
      msg += f"{padding}style {self.id} stroke:#0ff\n"

    for child in self.children:
      msg += child.mermaid(offset + 4, selected)
      msg += f"{padding}{self.id} --> {child.id}\n"

    return msg


class MCTS:
  question: str
  root: MCTSNode
  llm: 'llm.LLM'
  selected: Optional['ch.ChatNode']
  exploration_weight: float
  thoughts: int

  def __init__(self, **kwargs):
    self.question = kwargs.get("question")
    self.root = kwargs.get("root")
    self.llm = kwargs.get("llm")
    self.selected = None
    self.exploration_weight = kwargs.get(
      "exploration_weight", MCTS_EXPLORATION_CONSTANT.value
    )
    self.thoughts = kwargs.get("thoughts", MCTS_THOUGHTS.value)

  async def select(self):
    logger.debug("Selecting node...")
    node = self.root
    while node.children:
      node = self.uct_select(node)
    return node

  async def expand(self, node):
    logger.debug(f"Expanding node {node.id}...")
    await self.llm.emit_status(f"Thinking about {node.id}...")

    for _ in range(random.randint(self.thoughts, self.thoughts + 1)):
      thought = await self.generate_thought(node.content)
      await self.llm.emit_message(f"\nThought: {thought}\n")
      new_content = await self.update_approach(node.content, thought)
      child = self.create_node(content=new_content, parent=node)
      node.add_child(child)

    return random.choice(node.children)

  async def simulate(self, node: MCTSNode):
    logger.debug(f"Simulating node {node.id}...")
    await self.llm.emit_status(f"Thinking about {node.id}...")
    await self.llm.emit_message(self.mermaid())
    return await self.evaluate_answer(node.content)

  def backpropagate(self, node: MCTSNode, score: float):
    logger.debug(f"Backpropagating from {node.id}...")
    while node:
      node.visits += 1
      node.value += score
      node = node.parent

  def uct_select(self, node: MCTSNode):
    logger.debug(f"Selecting uct {node.id}...")
    return max(node.children, key=lambda child: child.uct_value())

  def best_child(self):
    return self.root.best_child()

  async def search(self, num_simulations):
    logger.debug("Starting search...")

    for _ in range(num_simulations):
      leaf = await self.select()
      self.selected = leaf
      if not leaf.fully_expanded():
        leaf = await self.expand(leaf)
      score = await self.simulate(leaf)
      self.backpropagate(leaf, score)

    return self.selected

  def create_node(self, **kwargs):
    node = MCTSNode(**kwargs)
    node.exploration_weight = self.exploration_weight

    return node

  async def generate_thought(self, answer):
    return await self.llm.chat_completion(
      prompt=thoughts_prompt,
      answer=answer,
      question=self.question,
      resolve=True
    )

  async def analyze_iteration(self, best_answer, best_score):
    return await self.llm.chat_completion(
      prompt=analyze_prompt,
      question=self.question,
      best_answer=best_answer,
      best_score=best_score,
      resolve=True
    )

  async def update_approach(self, answer, improvements):
    return await self.llm.chat_completion(
      prompt=update_prompt,
      question=self.question,
      answer=answer,
      improvements=improvements,
      resolve=True,
    )

  async def evaluate_answer(self, answer):
    result = await self.llm.chat_completion(
      prompt=eval_answer_prompt,
      answer=answer,
      question=self.question,
      resolve=True,
    )

    try:
      score = re.search(r"\d+", result).group()
      return int(score)
    except AttributeError:
      logger.error(f"AnswerEval: unable to parse \"{result[:100]}\"")
      return 0

  def mermaid(self, selected=None):
    return f"""
```mermaid
graph LR
{self.root.mermaid(0, selected.id if selected else self.selected.id)}
```
"""


def escape_mermaid(text):
  return text.replace('"', "&quot;").replace("(", "&#40;").replace(")", "&#41;")


# ==============================================================================
async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  strat = MCTS_STRAT.value
  strat_params = MCTS_STRAT_PARAMS.value
  exploration_constant = MCTS_EXPLORATION_CONSTANT.value
  max_simulations = MCTS_MAX_SIMULATIONS.value
  max_iterations = MCTS_MAX_ITERATIONS.value
  thoughts = MCTS_THOUGHTS.value

  debug_info = {
    "strat": strat,
    "strat_params": strat_params,
    "exploration_constant": exploration_constant,
    "max_simulations": max_simulations,
    "max_iterations": max_iterations,
    "thoughts": thoughts,
  }

  logger.debug(f"{ID_PREFIX}: {debug_info}")
  nodes = selection.apply_strategy(chat, strategy=strat, params=strat_params)

  if (len(nodes) > 1):
    logger.warning(
      f"{ID_PREFIX}: Matched multiple nodes, only the first one will be processed."
    )

  if len(nodes) == 0:
    log.info(f"{ID_PREFIX}: No nodes matched, skipping.")
    return llm.stream_chat_completion()

  node = nodes[0]
  question = node.content

  await llm.emit_status('Preparing initial thoughts...')
  mcts_chat = ch.Chat(
    llm=llm,
    tail=MCTSNode(
      role="user", content=initial_prompt.format(question=question)
    )
  )
  mcts_chat.chat_node_type = MCTSNode
  mcts_chat.llm = llm
  await mcts_chat.emit_advance()

  await llm.emit_status('Starting MCTS search...')
  mcts = MCTS(
    question=question,
    root=mcts_chat.tail,
    llm=llm,
    exploration_weight=exploration_constant,
    thoughts=thoughts
  )

  best_answer = None
  best_score = -float("inf")

  for i in range(max_iterations):
    await llm.emit_status(f"MCTS iteration {i + 1}/{max_iterations}...")
    best_node = await mcts.search(max_simulations)
    score = await mcts.evaluate_answer(best_node.content)

    if score > best_score:
      best_answer = best_node.content
      best_score = score

  # Final completion
  mcts_chat.assistant(f"Here is the best answer I can think of: {best_answer}")
  mcts_chat.user('Thank you, now please summarize it for me.')
  await llm.stream_final_completion(chat=mcts_chat)



"""
This module is responsible for loading all the modules
from the configured folders and registering them in
the registry.
"""

import os
import importlib

from config import INTERMEDIATE_OUTPUT, BOOST_FOLDERS

import log

logger = log.setup_logger(__name__)

registry = {}


def load_folder(folder):
  logger.debug(f"Loading modules from '{folder}'")
  for filename in os.listdir(folder):
    is_target_mod = filename.endswith(".py") and filename != "__init__.py"

    if is_target_mod:
      module_name = filename[:-3]
      module = importlib.import_module(f"{folder}.{module_name}")

      if hasattr(module, "ID_PREFIX"):
        logger.debug(f"Registering '{module.ID_PREFIX}'...")
        registry[module_name] = module


for folder in BOOST_FOLDERS.value:
  load_folder(folder)

if len(registry) == 0:
  logger.warning("No modules loaded. Is boost configured correctly?")
else:
  logger.info(f"Loaded {len(registry)} modules: {', '.join(registry.keys())}")



# Monkey-patch to include "/v1/models" endpoint to make it complatible
# with the Open WebUI

import json
from time import time
from uuid import uuid4
from fastapi import APIRouter
from fastapi.requests import Request
from ktransformers.server.utils.create_interface import get_interface
from ktransformers.server.schemas.assistants.streaming import chat_stream_response
from ktransformers.server.schemas.endpoints.chat import ChatCompletionCreate,ChatCompletionChunk,ChatCompletionObject
from ktransformers.server.backend.base import BackendInterfaceBase

router = APIRouter()

@router.get('/models', tags=['openai'])
async def models():
    return {
        "object": "list",
        "data": [
            {
                "id": "ktransformers",
                "object": "model",
                "created": 1234567890,
                "owned_by": "organization",
                "permission": []
            }
        ]
    }

@router.post('/chat/completions',tags=['openai'])
async def chat_completion(request:Request,create:ChatCompletionCreate):
    id = str(uuid4())

    interface: BackendInterfaceBase = get_interface()
    # input_ids = interface.format_and_tokenize_input_ids(id,messages=create.get_tokenizer_messages())

    input_message = [json.loads(m.model_dump_json()) for m in create.messages]

    if create.stream:
        async def inner():
            chunk = ChatCompletionChunk(id=id,object='chat.completion.chunk',created=int(time()))
            async for token in interface.inference(input_message,id):
                chunk.set_token(token)
                yield chunk
        return chat_stream_response(request,inner())
    else:
        comp = ChatCompletionObject(id=id,object='chat.completion.chunk',created=int(time()))
        async for token in interface.inference(input_message,id):
            comp.append_token(token)
        return comp



import random

import chat as ch
import llm
import log

ID_PREFIX = 'cea'    # 'cellular automata'


def cellular_automata(rule, initial_state, generations):
  """
    Runs a one-dimensional cellular automata and records results in binary,
    allowing the state to grow.

    Args:
        rule (int): The rule number for the cellular automata (0-255).
        initial_state (list): The initial state of the cellular automata.
        generations (int): The number of generations to run.

    Returns:
        list: A list of binary strings representing the state of the cellular automata at each generation.
    """
  # Convert the rule number to a binary string and pad with zeros to 8 bits
  rule_binary = format(rule, '08b')

  # Initialize the list to store the results
  results = ["".join(map(str, initial_state))]

  # Run the cellular automata for the specified number of generations
  current_state = initial_state.copy()
  for _ in range(generations):
    # Initialize the next state with a zero on each end
    next_state = [0] + current_state + [0]

    # Apply the rule to each cell in the current state
    for i in range(1, len(next_state) - 1):
      # Get the left, center, and right cells
      left = current_state[i - 2] if i > 1 else 0
      center = current_state[i - 1]
      right = current_state[i] if i < len(current_state) else 0

      # Convert the left, center, and right cells to a binary string
      neighborhood = f"{left}{center}{right}"

      # Get the next state of the cell based on the rule
      next_state[i] = int(rule_binary[7 - int(neighborhood, 2)])

    # Update the current state and append the next state to the results
    current_state = next_state
    results.append("".join(map(str, next_state)))

  return results

def render_ca(results):
  """
    Renders the results of a cellular automata as a string.

    Args:
        results (list): A list of binary strings representing the state of the cellular automata at each generation.

    Returns:
        str: A string representation of the cellular automata results.
    """
  return join.join(["".join(["|" if cell == "1" else "." for cell in result]) for result in results])



initial_state = [1]
join = '\n'


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  rule = int(llm.boost_params.get('cea_rule', '73'))
  gens = int(llm.boost_params.get('cea_generations', '32'))

  chat.user(
    f"""
Before completing my request, please think for a while.
    """.strip()
  )
  chat.assistant(
    f"""Good idea! Let me think...

```thoughts
{render_ca(cellular_automata(rule, initial_state, gens))}
```

"""
  )
  await llm.emit_message(chat.tail.content)
  chat.user(f"""
Now, please address my request.
    """.strip())
  await llm.stream_final_completion()



import llm
import chat as ch

ID_PREFIX = 'l33t'

async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  side_chat = ch.Chat.from_conversation([chat.tail.message()])
  side_chat.llm = llm

  await llm.emit_status('l33t speak...')
  side_chat.user("Rewrite my previous message in light leetspeak, ensure it is still readable by an average user, though.")
  await side_chat.emit_advance()
  await llm.emit_status('l33t answer...')
  side_chat.user(
    "Now write an answer for that message, also use l33t speak, but not too much. Refer to original message for clarity."
  )
  await side_chat.emit_advance()
  await llm.emit_status('Unl33t...')
  side_chat.user('Now write a final answer, in plain English')
  await side_chat.emit_advance()


import llm
import log
import chat as ch
import config
import selection

ID_PREFIX = "eli5"
logger = log.setup_logger(ID_PREFIX)

eli5_prompt = """
My friend asked me this question: "{question}".
Explain it to me like I'm stupid. Explain every word and its specific impact on the question.
Do not asnwer the question, though, I want to figure it out myself.
I just need a simpler explanation thats easy to understand and follow.
""".strip()

answer_prompt = """
<instruction>
Given the initial question and its dedetailed explanation, provide the answer to the question.
</instruction>

<question>
{question}
</question>

<explanation>
{explanation}
</explanation>
""".strip()


async def apply(chat: 'ch.Chat', llm: 'llm.LLM'):
  strat = config.ELI5_STRAT.value
  strat_params = config.ELI5_STRAT_PARAMS.value
  debug_info = {
    "strat": strat,
    "strat_params": strat_params,
  }
  logger.debug(f"{ID_PREFIX}: {debug_info}")

  nodes = selection.apply_strategy(chat, strategy=strat, params=strat_params)

  if (len(nodes) > 1):
    logger.warning(
      f"{ID_PREFIX}: Matched multiple nodes, only the first one will be processed."
    )

  if len(nodes) == 0:
    log.info(f"{ID_PREFIX}: No nodes matched, skipping.")
    return await llm.stream_final_completion()

  node = nodes[0]
  question = node.content

  await llm.emit_status("Explaining the question to myself...")
  explanation = await llm.stream_chat_completion(
    prompt=eli5_prompt.format(question=question),
  )

  await llm.emit_status('ELI5 Response')
  await llm.stream_final_completion(
    prompt=answer_prompt.format(
      question=question,
      explanation=explanation,
    )
  )
